import pytest
from quart.exceptions import RequestEntityTooLarge

from quart_trio.request import TrioBody


@pytest.mark.trio
async def test_body_exceeds_max_content_length() -> None:
    max_content_length = 5
    body = TrioBody(None, max_content_length)
    body.append(b" " * (max_content_length + 1))
    with pytest.raises(RequestEntityTooLarge):
        await body



import threading
from typing import Generator

import pytest
from quart import request, ResponseReturnValue

from quart_trio import QuartTrio


@pytest.fixture(name="app")
def _app() -> QuartTrio:
    app = QuartTrio(__name__)

    @app.route("/", methods=["GET", "POST"])
    def index() -> ResponseReturnValue:
        return request.method

    @app.route("/gen")
    def gen() -> ResponseReturnValue:
        def _gen() -> Generator[bytes, None, None]:
            yield b"%d" % threading.current_thread().ident
            for _ in range(2):
                yield b"b"

        return _gen(), 200

    return app


@pytest.mark.trio
async def test_sync_request_context(app: QuartTrio) -> None:
    test_client = app.test_client()
    response = await test_client.get("/")
    assert b"GET" in (await response.get_data())  # type: ignore
    response = await test_client.post("/")
    assert b"POST" in (await response.get_data())  # type: ignore


@pytest.mark.trio
async def test_sync_generator(app: QuartTrio) -> None:
    test_client = app.test_client()
    response = await test_client.get("/gen")
    result = await response.get_data()
    assert result[-2:] == b"bb"  # type: ignore
    assert int(result[:-2]) != threading.current_thread().ident



from contextlib import asynccontextmanager
from typing import AnyStr, AsyncGenerator, List, Optional, Union

import trio
from quart.exceptions import BadRequest
from quart.testing import make_test_headers_path_and_query_string, QuartClient, WebsocketResponse
from quart.wrappers import Request, Response, Websocket
from werkzeug.datastructures import Headers
from werkzeug.exceptions import BadRequest as WBadRequest


class _TestingWebsocket:
    def __init__(
        self,
        server_send: trio._channel.MemorySendChannel,
        client_receive: trio._channel.MemoryReceiveChannel,
    ) -> None:
        self.server_send = server_send
        self.client_receive = client_receive
        self.accepted = False

    async def receive(self) -> AnyStr:
        return await self.client_receive.receive()

    async def send(self, data: AnyStr) -> None:
        await self.server_send.send(data)

    async def accept(self, headers: Headers, subprotocol: Optional[str]) -> None:
        self.accepted = True
        self.accept_headers = headers
        self.accept_subprotocol = subprotocol


class TrioQuartClient(QuartClient):
    async def _handle_request(self, request: Request) -> Response:
        return await self.app.handle_request(request)

    @asynccontextmanager
    async def websocket(  # type: ignore
        self,
        path: str,
        *,
        headers: Optional[Union[dict, Headers]] = None,
        query_string: Optional[dict] = None,
        scheme: str = "ws",
        subprotocols: Optional[List[str]] = None,
        root_path: str = "",
        http_version: str = "1.1",
    ) -> AsyncGenerator[_TestingWebsocket, None]:
        headers, path, query_string_bytes = make_test_headers_path_and_query_string(
            self.app, path, headers, query_string
        )
        server_send, server_receive = trio.open_memory_channel(10)
        client_send, client_receive = trio.open_memory_channel(10)
        websocket_client = _TestingWebsocket(server_send, client_receive)

        websocket = self.app.websocket_class(
            path,
            query_string_bytes,
            scheme,
            headers,
            root_path,
            http_version,
            subprotocols,
            server_receive.receive,
            client_send.send,
            websocket_client.accept,
        )
        adapter = self.app.create_url_adapter(websocket)
        try:
            url_rule, _ = adapter.match()
        except WBadRequest:
            raise BadRequest()

        async with trio.open_nursery() as nursery:
            nursery.start_soon(self._handle_websocket, websocket)
            try:
                yield websocket_client
            finally:
                nursery.cancel_scope.cancel()

    async def _handle_websocket(self, websocket: Websocket) -> None:
        response = await self.app.handle_websocket(websocket)
        if response is not None:
            raise WebsocketResponse(response)



import warnings
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Awaitable, Callable, Coroutine, Optional, Union

import trio
from hypercorn.config import Config as HyperConfig
from hypercorn.trio import serve
from quart import Quart, request_started, websocket_started
from quart.ctx import RequestContext, WebsocketContext
from quart.logging import create_serving_logger
from quart.utils import is_coroutine_function
from quart.wrappers import Request, Response, Websocket

from .asgi import TrioASGIHTTPConnection, TrioASGILifespan, TrioASGIWebsocketConnection
from .request import TrioRequest, TrioWebsocket
from .response import TrioResponse
from .testing import TrioQuartClient
from .utils import run_sync


class QuartTrio(Quart):
    nursery: trio._core.Nursery
    asgi_http_class = TrioASGIHTTPConnection
    asgi_lifespan_class = TrioASGILifespan  # type: ignore
    asgi_websocket_class = TrioASGIWebsocketConnection
    lock_class = trio.Lock
    request_class = TrioRequest
    response_class = TrioResponse
    test_client_class = TrioQuartClient
    websocket_class = TrioWebsocket

    def run(  # type: ignore
        self,
        host: str = "127.0.0.1",
        port: int = 5000,
        debug: Optional[bool] = None,
        use_reloader: bool = True,
        ca_certs: Optional[str] = None,
        certfile: Optional[str] = None,
        keyfile: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Run this application.

        This is best used for development only, see using Gunicorn for
        production servers.

        Arguments:
            host: Hostname to listen on. By default this is loopback
                only, use 0.0.0.0 to have the server listen externally.
            port: Port number to listen on.
            debug: If set enable (or disable) debug mode and debug output.
            use_reloader: Automatically reload on code changes.
            ca_certs: Path to the SSL CA certificate file.
            certfile: Path to the SSL certificate file.
            ciphers: Ciphers to use for the SSL setup.
            keyfile: Path to the SSL key file.
        """
        if kwargs:
            warnings.warn(
                f"Additional arguments, {','.join(kwargs.keys())}, are not supported.\n"
                "They may be supported by Hypercorn, which is the ASGI server Quart "
                "uses by default. This method is meant for development and debugging."
            )

        scheme = "https" if certfile is not None and keyfile is not None else "http"
        print(f"Running on {scheme}://{host}:{port} (CTRL + C to quit)")  # noqa: T001, T002

        trio.run(self.run_task, host, port, debug, use_reloader, ca_certs, certfile, keyfile)

    def run_task(
        self,
        host: str = "127.0.0.1",
        port: int = 5000,
        debug: Optional[bool] = None,
        use_reloader: bool = True,
        ca_certs: Optional[str] = None,
        certfile: Optional[str] = None,
        keyfile: Optional[str] = None,
        shutdown_trigger: Optional[Callable[..., Awaitable[None]]] = None,
    ) -> Coroutine[None, None, None]:
        """Return a task that when awaited runs this application.

        This is best used for development only, see Hypercorn for
        production servers.

        Arguments:
            host: Hostname to listen on. By default this is loopback
                only, use 0.0.0.0 to have the server listen externally.
            port: Port number to listen on.
            debug: If set enable (or disable) debug mode and debug output.
            use_reloader: Automatically reload on code changes.
            loop: Asyncio loop to create the server in, if None, take default one.
                If specified it is the caller's responsibility to close and cleanup the
                loop.
            ca_certs: Path to the SSL CA certificate file.
            certfile: Path to the SSL certificate file.
            keyfile: Path to the SSL key file.

        """
        config = HyperConfig()
        config.access_log_format = "%(h)s %(r)s %(s)s %(b)s %(D)s"
        config.accesslog = create_serving_logger()
        config.bind = [f"{host}:{port}"]
        config.ca_certs = ca_certs
        config.certfile = certfile
        if debug is not None:
            config.debug = debug
        config.errorlog = config.accesslog
        config.keyfile = keyfile
        config.use_reloader = use_reloader

        return serve(self, config)

    def ensure_async(self, func: Callable[..., Any]) -> Callable[..., Awaitable[Any]]:
        """Ensure that the returned func is async and calls the func.

        .. versionadded:: 0.11

        Override if you wish to change how synchronous functions are
        run. Before Quart 0.11 this did not run the synchronous code
        in an executor.
        """
        if is_coroutine_function(func):
            return func
        else:
            return run_sync(func)

    @asynccontextmanager
    async def test_app(self) -> AsyncGenerator["Quart", None]:
        async with trio.open_nursery() as nursery:
            self.nursery = nursery
            await self.startup()
            try:
                yield self
            finally:
                await self.shutdown()
        self.nursery = None

    async def handle_request(self, request: Request, *, _preserve: bool = False) -> Response:
        async with self.request_context(request, _preserve=_preserve) as request_context:
            try:
                return await self.full_dispatch_request(request_context)
            except trio.Cancelled:
                raise  # Cancelled should be handled by serving code.
            except (Exception, trio.MultiError) as error:
                return await self.handle_exception(error)

    async def full_dispatch_request(
        self, request_context: Optional[RequestContext] = None
    ) -> Response:
        """Adds pre and post processing to the request dispatching.

        Arguments:
            request_context: The request context, optional as Flask
                omits this argument.
        """
        await self.try_trigger_before_first_request_functions()
        await request_started.send(self)
        try:
            result = await self.preprocess_request(request_context)
            if result is None:
                result = await self.dispatch_request(request_context)
        except (Exception, trio.MultiError) as error:
            result = await self.handle_user_exception(error)
        return await self.finalize_request(result, request_context)

    async def handle_user_exception(self, error: Union[Exception, trio.MultiError]) -> Response:
        if isinstance(error, trio.MultiError):
            for exception in error.exceptions:
                try:
                    return await self.handle_user_exception(exception)
                except Exception:
                    pass  # No handler for this error
            # Not found a single handler, re-raise the error
            raise error
        else:
            return await super().handle_user_exception(error)

    async def handle_websocket(
        self, websocket: Websocket, *, _preserve: bool = False
    ) -> Optional[Response]:
        async with self.websocket_context(websocket, _preserve=_preserve) as websocket_context:
            try:
                return await self.full_dispatch_websocket(websocket_context)
            except trio.Cancelled:
                raise  # Cancelled should be handled by serving code.
            except (Exception, trio.MultiError) as error:
                return await self.handle_websocket_exception(error)

    async def full_dispatch_websocket(
        self, websocket_context: Optional[WebsocketContext] = None
    ) -> Optional[Response]:
        """Adds pre and post processing to the websocket dispatching.

        Arguments:
            websocket_context: The websocket context, optional to match
                the Flask convention.
        """
        await self.try_trigger_before_first_request_functions()
        await websocket_started.send(self)
        try:
            result = await self.preprocess_websocket(websocket_context)
            if result is None:
                result = await self.dispatch_websocket(websocket_context)
        except (Exception, trio.MultiError) as error:
            result = await self.handle_user_exception(error)
        return await self.finalize_websocket(result, websocket_context)



from contextvars import copy_context
from functools import partial, wraps
from inspect import isgenerator
from typing import Any, AsyncGenerator, Callable, Coroutine, Generator

import trio


def run_sync(func: Callable[..., Any]) -> Callable[..., Coroutine[Any, None, None]]:
    """Ensure that the sync function is run without blocking.

    If the *func* is not a coroutine it will be wrapped such that it
    runs on a thread. This ensures that synchronous functions do not
    block the event loop.
    """

    @wraps(func)
    async def _wrapper(*args: Any, **kwargs: Any) -> Any:
        result = await trio.to_thread.run_sync(
            partial(copy_context().run, partial(func, *args, **kwargs))
        )
        if isgenerator(result):
            return run_sync_iterable(result)
        else:
            return result

    _wrapper._quart_async_wrapper = True  # type: ignore
    return _wrapper


def run_sync_iterable(iterable: Generator[Any, None, None]) -> AsyncGenerator[Any, None]:
    async def _gen_wrapper() -> AsyncGenerator[Any, None]:
        # Wrap the generator such that each iteration runs
        # in the executor. Then rationalise the raised
        # errors so that it ends.
        def _inner() -> Any:
            # https://bugs.python.org/issue26221
            # StopIteration errors are swallowed by the
            # run_in_exector method
            try:
                return next(iterable)
            except StopIteration:
                raise StopAsyncIteration()

        while True:
            try:
                yield await trio.to_thread.run_sync(partial(copy_context().run, _inner))
            except StopAsyncIteration:
                return

    return _gen_wrapper()



import pytest
import trio

from quart_trio.app import QuartTrio
from quart_trio.asgi import TrioASGIWebsocketConnection


@pytest.mark.trio
async def test_websocket_complete_on_disconnect() -> None:
    scope = {
        "asgi": {"version": "3"},
        "headers": [(b"host", b"quart")],
        "http_version": "1.1",
        "method": "GET",
        "scheme": "wss",
        "path": "ws://quart/path",
        "query_string": b"",
        "subprotocols": [],
        "extensions": {"websocket.http.response": {}},
    }
    connection = TrioASGIWebsocketConnection(QuartTrio(__name__), scope)
    send_channel, receive_channel = trio.open_memory_channel(0)
    async with trio.open_nursery() as nursery:
        nursery.start_soon(connection.handle_messages, nursery, receive_channel.receive)
        await send_channel.send({"type": "websocket.disconnect"})
    assert nursery.cancel_scope.cancelled_caught



from functools import partial
from typing import Callable, cast, Optional, TYPE_CHECKING

import trio
from quart.asgi import ASGIHTTPConnection, ASGIWebsocketConnection
from quart.wrappers import Request, Response, sentinel, Websocket  # noqa: F401
from werkzeug.datastructures import Headers

if TYPE_CHECKING:
    from quart_trio import QuartTrio  # noqa: F401


class TrioASGIHTTPConnection(ASGIHTTPConnection):
    async def __call__(self, receive: Callable, send: Callable) -> None:
        request = self._create_request_from_scope(send)
        async with trio.open_nursery() as nursery:
            nursery.start_soon(self.handle_messages, nursery, request, receive)
            nursery.start_soon(self.handle_request, nursery, request, send)

    async def handle_messages(  # type: ignore
        self, nursery: trio._core._run.Nursery, request: Request, receive: Callable
    ) -> None:
        await super().handle_messages(request, receive)
        nursery.cancel_scope.cancel()

    async def handle_request(  # type: ignore
        self, nursery: trio._core._run.Nursery, request: Request, send: Callable
    ) -> None:
        response = await self.app.handle_request(request)
        if response.timeout != sentinel:
            timeout = cast(Optional[float], response.timeout)
        else:
            timeout = self.app.config["RESPONSE_TIMEOUT"]

        if timeout is not None:
            with trio.move_on_after(timeout):
                await self._send_response(send, response)
        else:
            await self._send_response(send, response)
        nursery.cancel_scope.cancel()


class TrioASGIWebsocketConnection(ASGIWebsocketConnection):
    def __init__(self, app: "QuartTrio", scope: dict) -> None:
        super().__init__(app, scope)
        self.send_channel, self.receive_channel = trio.open_memory_channel(10)

    async def __call__(self, receive: Callable, send: Callable) -> None:
        websocket = self._create_websocket_from_scope(send)
        async with trio.open_nursery() as nursery:
            nursery.start_soon(self.handle_messages, nursery, receive)
            nursery.start_soon(self.handle_websocket, nursery, websocket, send)

    async def handle_messages(  # type: ignore
        self, nursery: trio._core._run.Nursery, receive: Callable
    ) -> None:
        while True:
            event = await receive()
            if event["type"] == "websocket.receive":
                await self.send_channel.send(event.get("bytes") or event["text"])
            elif event["type"] == "websocket.disconnect":
                break
        nursery.cancel_scope.cancel()

    def _create_websocket_from_scope(self, send: Callable) -> Websocket:
        headers = Headers()
        headers["Remote-Addr"] = (self.scope.get("client") or ["<local>"])[0]
        for name, value in self.scope["headers"]:
            headers.add(name.decode().title(), value.decode())

        return self.app.websocket_class(
            self.scope["path"],
            self.scope["query_string"],
            self.scope["scheme"],
            headers,
            self.scope.get("root_path", ""),
            self.scope.get("http_version", "1.1"),
            self.scope.get("subprotocols", []),
            self.receive_channel.receive,
            partial(self.send_data, send),
            partial(self.accept_connection, send),
        )

    async def handle_websocket(  # type: ignore
        self, nursery: trio._core._run.Nursery, websocket: Websocket, send: Callable
    ) -> None:
        await super().handle_websocket(websocket, send)
        nursery.cancel_scope.cancel()


class TrioASGILifespan:
    def __init__(self, app: "QuartTrio", scope: dict) -> None:
        self.app = app

    async def __call__(self, receive: Callable, send: Callable) -> None:
        async with trio.open_nursery() as nursery:
            self.app.nursery = nursery
            while True:
                event = await receive()
                if event["type"] == "lifespan.startup":
                    try:
                        await self.app.startup()
                    except Exception as error:
                        await send({"type": "lifespan.startup.failed", "message": str(error)})
                    else:
                        await send({"type": "lifespan.startup.complete"})
                elif event["type"] == "lifespan.shutdown":
                    try:
                        await self.app.shutdown()
                    except Exception as error:
                        await send({"type": "lifespan.shutdown.failed", "message": str(error)})
                    else:
                        await send({"type": "lifespan.shutdown.complete"})
                    break



from pathlib import Path

import pytest
from py._path.local import LocalPath
from quart import abort, Quart, ResponseReturnValue, send_file, websocket
from quart.testing import WebsocketResponse

from quart_trio import QuartTrio


@pytest.fixture
def app() -> Quart:
    app = QuartTrio(__name__)

    @app.route("/")
    async def index() -> ResponseReturnValue:
        return "index"

    @app.websocket("/ws/")
    async def ws() -> None:
        # async for message in websocket:
        while True:
            message = await websocket.receive()
            await websocket.send(message)

    @app.websocket("/ws/abort/")
    async def ws_abort() -> None:
        abort(401)

    return app


@pytest.mark.trio
async def test_index(app: Quart) -> None:
    test_client = app.test_client()
    response = await test_client.get("/")
    assert response.status_code == 200
    assert b"index" in (await response.get_data())  # type: ignore


@pytest.mark.trio
async def test_websocket(app: Quart) -> None:
    test_client = app.test_client()
    data = b"bob"
    async with test_client.websocket("/ws/") as test_websocket:
        await test_websocket.send(data)
        result = await test_websocket.receive()
    assert result == data  # type: ignore


@pytest.mark.trio
async def test_websocket_abort(app: Quart) -> None:
    test_client = app.test_client()
    try:
        async with test_client.websocket("/ws/abort/") as test_websocket:
            await test_websocket.receive()
    except WebsocketResponse as error:
        assert error.response.status_code == 401


@pytest.mark.trio
async def test_send_file_path(tmpdir: LocalPath) -> None:
    app = QuartTrio(__name__)
    file_ = tmpdir.join("send.img")
    file_.write("something")
    async with app.app_context():
        response = await send_file(Path(file_.realpath()))
    assert (await response.get_data(raw=True)) == file_.read_binary()



from typing import NoReturn

import pytest
import trio
from quart import ResponseReturnValue
from quart.testing import WebsocketResponse

from quart_trio import QuartTrio


@pytest.fixture(name="error_app", scope="function")
def _error_app() -> QuartTrio:
    app = QuartTrio(__name__)

    @app.route("/")
    async def index() -> NoReturn:
        raise trio.MultiError([ValueError(), trio.MultiError([TypeError(), ValueError()])])

    @app.websocket("/ws/")
    async def ws() -> NoReturn:
        raise trio.MultiError([ValueError(), trio.MultiError([TypeError(), ValueError()])])

    return app


@pytest.mark.trio
async def test_multi_error_handling(error_app: QuartTrio) -> None:
    @error_app.errorhandler(TypeError)
    async def handler(_: Exception) -> ResponseReturnValue:
        return "", 201

    test_client = error_app.test_client()
    response = await test_client.get("/")
    assert response.status_code == 201


@pytest.mark.trio
async def test_websocket_multi_error_handling(error_app: QuartTrio) -> None:
    @error_app.errorhandler(TypeError)
    async def handler(_: Exception) -> ResponseReturnValue:
        return "", 201

    test_client = error_app.test_client()
    try:
        async with test_client.websocket("/ws/") as test_websocket:
            await test_websocket.receive()
    except WebsocketResponse as error:
        assert error.response.status_code == 201


@pytest.mark.trio
async def test_multi_error_unhandled(error_app: QuartTrio) -> None:
    test_client = error_app.test_client()
    response = await test_client.get("/")
    assert response.status_code == 500


@pytest.mark.trio
async def test_websocket_multi_error_unhandled(error_app: QuartTrio) -> None:
    test_client = error_app.test_client()
    try:
        async with test_client.websocket("/ws/") as test_websocket:
            await test_websocket.receive()
    except WebsocketResponse as error:
        assert error.response.status_code == 500


@pytest.mark.trio
async def test_test_app() -> None:
    startup = False
    shutdown = False

    app = QuartTrio(__name__)

    @app.before_serving
    async def before() -> None:
        nonlocal startup
        startup = True

    @app.after_serving
    async def after() -> None:
        nonlocal shutdown
        shutdown = True

    @app.route("/")
    async def index() -> str:
        return ""

    async with app.test_app() as test_app:
        assert startup
        assert app.nursery is not None
        test_client = test_app.test_client()
        await test_client.get("/")
        assert not shutdown
    assert shutdown



from typing import AnyStr, Optional

import trio
from quart.exceptions import RequestEntityTooLarge, RequestTimeout
from quart.wrappers.request import Body, Request, Websocket


class EventWrapper:
    def __init__(self) -> None:
        self._event = trio.Event()

    def clear(self) -> None:
        self._event = trio.Event()

    async def wait(self) -> None:
        await self._event.wait()

    def set(self) -> None:
        self._event.set()


class TrioBody(Body):
    def __init__(
        self, expected_content_length: Optional[int], max_content_length: Optional[int]
    ) -> None:
        self._data = bytearray()
        self._complete = EventWrapper()  # type: ignore
        self._has_data = EventWrapper()  # type: ignore
        self._max_content_length = max_content_length
        # Exceptions must be raised within application (not ASGI)
        # calls, this is achieved by having the ASGI methods set this
        # to an exception on error.
        self._must_raise: Optional[Exception] = None
        if (
            expected_content_length is not None
            and max_content_length is not None
            and expected_content_length > max_content_length
        ):
            self._must_raise = RequestEntityTooLarge()


class TrioRequest(Request):
    body_class = TrioBody

    async def get_data(self, raw: bool = True) -> AnyStr:
        if self.body_timeout is not None:
            with trio.move_on_after(self.body_timeout) as cancel_scope:
                return await self.body
        else:
            return await self.body
        if cancel_scope.cancelled_caught:
            raise RequestTimeout()


class TrioWebsocket(Websocket):
    async def send(self, data: AnyStr) -> None:
        await self.accept()
        await self._send(data)



import os

from setuptools import setup, find_packages

PROJECT_ROOT = os.path.dirname(__file__)

with open(os.path.join(PROJECT_ROOT, 'README.rst')) as file_:
    long_description = file_.read()

INSTALL_REQUIRES = [
    'hypercorn[trio] >= 0.6.0',
    'quart >= 0.12.0',
    'trio >= 0.9.0',
]

setup(
    name='Quart-Trio',
    version='0.6.0',
    python_requires='>=3.7.0',
    description="A Quart extension to provide trio support.",
    long_description=long_description,
    url='https://gitlab.com/pgjones/quart-trio/',
    author='P G Jones',
    author_email='philip.graham.jones@googlemail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=['quart_trio'],
    install_requires=INSTALL_REQUIRES,
    tests_require=INSTALL_REQUIRES + [
        'pytest',
        'pytest-trio',
    ],
    include_package_data=True,
)



import os
from inspect import isasyncgen, isgenerator
from types import TracebackType
from typing import AsyncGenerator, Iterable, Optional, Union

import trio
from quart.wrappers.response import _raise_if_invalid_range, IterableBody, Response, ResponseBody

from .utils import run_sync_iterable


class TrioFileBody(ResponseBody):
    """Provides an async file accessor with range setting.

    The :attr:`Response.response` attribute must be async-iterable and
    yield bytes, which this wrapper does for a file. In addition it
    allows a range to be set on the file, thereby supporting
    conditional requests.
    """

    buffer_size = 8192

    def __init__(
        self, file_path: Union[str, bytes, os.PathLike], *, buffer_size: Optional[int] = None
    ) -> None:
        self.file_path = file_path
        self.size = os.path.getsize(self.file_path)
        self.begin = 0
        self.end = self.size
        if buffer_size is not None:
            self.buffer_size = buffer_size
        self.file: Optional[trio._file_io.AsyncIOWrapper] = None

    async def __aenter__(self) -> "TrioFileBody":
        self.file = await trio.open_file(self.file_path, mode="rb")
        await self.file.__aenter__()
        await self.file.seek(self.begin)
        return self

    async def __aexit__(self, exc_type: type, exc_value: BaseException, tb: TracebackType) -> None:
        await self.file.__aexit__(exc_type, exc_value, tb)

    def __aiter__(self) -> "TrioFileBody":
        return self

    async def __anext__(self) -> bytes:
        current = await self.file.tell()
        if current >= self.end:
            raise StopAsyncIteration()
        read_size = min(self.buffer_size, self.end - current)
        chunk = await self.file.read(read_size)

        if chunk:
            return chunk
        else:
            raise StopAsyncIteration()

    async def convert_to_sequence(self) -> bytes:
        result = bytearray()
        async with self as response:
            async for data in response:
                result.extend(data)
        return bytes(result)

    async def make_conditional(
        self, begin: int, end: Optional[int], max_partial_size: Optional[int] = None
    ) -> int:
        self.begin = begin
        self.end = self.size if end is None else end
        if max_partial_size is not None:
            self.end = min(self.begin + max_partial_size, self.end)
        _raise_if_invalid_range(self.begin, self.end, self.size)
        return self.size


class TrioIterableBody(IterableBody):
    def __init__(self, iterable: Union[AsyncGenerator[bytes, None], Iterable]) -> None:
        self.iter: AsyncGenerator[bytes, None]
        if isasyncgen(iterable):
            self.iter = iterable  # type: ignore
        elif isgenerator(iterable):
            self.iter = run_sync_iterable(iterable)  # type: ignore
        else:

            async def _aiter() -> AsyncGenerator[bytes, None]:
                for data in iterable:  # type: ignore
                    yield data

            self.iter = _aiter()


class TrioResponse(Response):
    file_body_class = TrioFileBody  # type: ignore
    iterable_body_class = TrioIterableBody
